"""
Test data loading
"""
import pandas as pd
from neighbors import load_movielens100k, load_toymat


def test_load_movielens100k():
    df = load_movielens100k()
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (100000, 4)


def test_load_toy_useritem_data():
    df = load_toymat()
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (50, 100)
    df = load_toymat(100, 50)
    assert df.shape == (100, 50)



"""
Modular testing for core algorithms. For ease of testing each test makes use of parameterized fixtures defined in conftest.py. Fixtures are passed in as args to each test function which automatically generates a complete grid of parameter combinations for each test.

Tests are split up by model for ease of modular testing (i.e. using pytest -k 'test_name') and to avoid creating uneccesary parameter combinations, thereby reducing the total number of tests.

For running tests in parallel `pip install pytest-xdist` and for nicer testing output `pip install pytest-sugar`.

Then you can run pytest locally using `pytest -rs -n auto`, to see skip messages at the end of the test session and visually confirm that only intended skipped tests are being skipped. To aid in this, all pytest.skip() messages end with 'OK' for intentionally skipped tests.
"""

from neighbors import Mean, KNN, NNMF_mult, NNMF_sgd, Base
import pytest
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def verify_fit(fit_kwargs):
    """Helper function to test fit call"""
    model = fit_kwargs.pop("model")
    model.fit(**fit_kwargs)
    assert model.is_fit
    assert model.predictions is not None
    return model.summary(), model


def verify_results(results, model, true_scores=None):
    """Helper function to test results object"""
    assert isinstance(results, pd.DataFrame)
    assert model.overall_results is not None
    assert model.user_results is not None
    assert model.user_results.shape == (model.data.shape[0], 4 * 2)
    # 4 metrics, 2 datasets (default .summary returns observed and missing only), 2 "groups" (all, subject)
    assert results.shape == (4 * 2 * 2, 5)
    if model.is_dense:
        assert not results.isnull().any().any()
    else:
        assert results.isnull().sum().sum() == 8

    # For regression testing our results should be relatively consistent, this just ensures that. We expect performances to be in the same ballpark (+/- 2), so ignore MSE because that fluctuates much more than RMSE, MAE, or corr
    if true_scores is not None:
        assert np.allclose(
            results.query("metric != 'mse' and group == 'all'").score.to_numpy(),
            true_scores,
            atol=2,
        )


def verify_transform(model):
    out = model.transform()
    # Handle edge case where masking leads to an entire column of NaNs i.e. no ratings at all for a single item
    if not len(
        model.masked_data.columns[model.masked_data.isnull().all()].tolist()
    ) and not len(model.predictions.columns[model.predictions.isnull().all()].tolist()):
        # Aside from that edge case there should be no missing values in the output of .transform()
        assert not out.isnull().any().any()
    else:
        pytest.skip("Skip masking edgecase - OK")

    out = model.transform(return_only_predictions=True)
    assert out.equals(model.predictions)


def verify_plotting(model):
    for dataset in ["full", "observed", "missing"]:
        out = model.plot_predictions(dataset=dataset)
        if dataset != "missing":
            assert isinstance(out, tuple)
        plt.close("all")


def test_downsample(simulate_wide_data):
    n_users, n_items = simulate_wide_data.shape
    cf = Base(simulate_wide_data)
    assert cf.data.shape == (n_users, n_items)

    # Test sampling_freq has no effect if target_type = 'samples'
    sampling_freq, target = 10, 2
    expected_items = int(n_items * (1 / target))
    cf.downsample(sampling_freq=sampling_freq, n_samples=target, target_type="samples")
    assert cf.data.shape == (n_users, expected_items)

    # Test each target_type
    sampling_freq, target = 10, 2
    cf = Base(simulate_wide_data)
    expected_items = int(n_items * (1 / (target * sampling_freq)))
    cf.downsample(sampling_freq=sampling_freq, n_samples=target, target_type="seconds")
    assert cf.data.shape == (n_users, expected_items)

    sampling_freq, target = 10, 5
    cf = Base(simulate_wide_data)
    expected_items = int(n_items * (1 / (sampling_freq / target)))
    cf.downsample(sampling_freq=sampling_freq, n_samples=target, target_type="hz")
    assert cf.data.shape == (n_users, expected_items)

    # Make sure downsampling affects fitted model artifacts
    cf = Mean(simulate_wide_data)
    cf.create_masked_data(n_mask_items=0.5)
    cf.fit()
    cf.downsample(sampling_freq=sampling_freq, n_samples=target, target_type="hz")
    assert cf.data.shape == (n_users, expected_items)
    assert cf.mask.shape == (n_users, expected_items)
    assert cf.masked_data.shape == (n_users, expected_items)
    assert cf.predictions.shape == (n_users, expected_items)

    # Make sure downsampling affects fitted model artifactsa including dilation
    cf = Mean(simulate_wide_data)
    cf.create_masked_data(n_mask_items=0.5)
    cf.fit(dilate_by_nsamples=5)
    cf.downsample(sampling_freq=sampling_freq, n_samples=target, target_type="hz")
    assert cf.data.shape == (n_users, expected_items)
    assert cf.mask.shape == (n_users, expected_items)
    assert cf.dilated_mask.shape == (n_users, expected_items)
    assert cf.masked_data.shape == (n_users, expected_items)
    assert cf.predictions.shape == (n_users, expected_items)


def test_init_and_dilate(init, mask, n_mask_items):
    """Test model initialization, initial masking, and dilation"""

    print(init.__class__.__name__)
    # Test that we calculate a mask
    if mask is not None or n_mask_items is not None:
        assert init.is_masked
        assert init.masked_data.isnull().any().any()

    # Test the mask is the right shape
    if n_mask_items is not None:
        total_items = init.data.shape[1]
        if isinstance(n_mask_items, (float, np.floating)):
            n_false_items = int(total_items * n_mask_items)
        else:
            n_false_items = n_mask_items
        calculated_n_false_items = init.masked_data.isnull().sum(1)[0]
        assert n_false_items == calculated_n_false_items

    # Test no accidental masking
    if mask is None and n_mask_items is None:
        assert not init.is_masked
        assert not init.masked_data.isnull().any().any()

        # Test fit failure when not masked
        with pytest.raises(ValueError):
            init.fit()
    if mask is not None or n_mask_items is not None:
        # Test dilation
        n_masked = init.masked_data.isnull().sum().sum()
        init.dilate_mask(n_samples=5)
        assert init.dilated_mask is not None
        assert init.is_mask_dilated is True
        # More values when we dilate the mask
        assert init.dilated_mask.sum().sum() > init.mask.sum().sum()
        # Fewer masked values after we dilate the mask
        assert n_masked > init.masked_data.isnull().sum().sum()


def test_mean(model, dilate_by_nsamples, n_mask_items):
    """Test Mean model"""
    if not isinstance(model, Mean):
        pytest.skip("Skip non Mean - OK")
    results, model = verify_fit(locals())
    if model.n_mask_items == 0.5 and not model.is_mask_dilated:
        true_scores = np.array(
            [
                0.55695942,
                19.73822567,
                25.1216878,
                0.61205836,
                18.62724324,
                23.57214148,
            ]
        )
    else:
        true_scores = None
    verify_results(results, model, true_scores)
    verify_plotting(model)
    verify_transform(model)


def test_knn(model, dilate_by_nsamples, n_mask_items, k, metric):
    """Test KNN model"""
    if not isinstance(model, KNN):
        pytest.skip("Skip non KNN - OK")
    results, model = verify_fit(locals())
    if (
        model.n_mask_items == 0.5
        and metric == "correlation"
        and not model.is_mask_dilated
        and k == 3
    ):
        true_scores = np.array(
            [
                0.84244861,
                14.40186486,
                17.49846264,
                0.86440771,
                13.20914251,
                15.95942564,
            ]
        )
    else:
        true_scores = None
    verify_results(results, model, true_scores)
    verify_plotting(model)
    verify_transform(model)


def test_nmf_mult(model, dilate_by_nsamples, n_mask_items, n_factors, n_iterations):
    """Test NNMF_mult model"""
    if not isinstance(model, NNMF_mult):
        pytest.skip("Skip non NNMF_mult - OK")
    results, model = verify_fit(locals())
    if (
        model.n_mask_items == 0.5
        and not model.is_mask_dilated
        and n_iterations == 100
        and n_factors is None
        and model.converged is True
    ):
        true_scores = np.array(
            [
                -2.78597027e-02,
                3.69112079e01,
                4.80280757e01,
                9.97852743e-01,
                1.01301480e00,
                1.93696252e00,
            ]
        )
    else:
        true_scores = None
    verify_results(results, model, true_scores)
    verify_plotting(model)
    verify_transform(model)
    # Smoke test for plotting learning curves
    model.plot_learning()
    plt.close("all")


def test_nmf_sgd(model, dilate_by_nsamples, n_mask_items, n_factors, n_iterations):
    """Test NNMF_sgd model"""
    if not isinstance(model, NNMF_sgd):
        pytest.skip("Skip non NNMF_sgd - OK")
    results, model = verify_fit(locals())
    if (
        model.n_mask_items == 0.5
        and not model.is_mask_dilated
        and n_iterations == 100
        and n_factors is None
        and model.converged is True
    ):
        true_scores = np.array(
            [
                0.78658875,
                15.44602272,
                18.85269805,
                0.99977349,
                0.49171737,
                0.63997854,
            ]
        )
    else:
        true_scores = None
    verify_results(results, model, true_scores)
    verify_plotting(model)
    verify_transform(model)
    # Smoke test for plotting learning curves
    model.plot_learning()
    plt.close("all")



"""
Included datasets
"""

import pandas as pd
import numpy as np
import pkg_resources
from .utils import check_random_state

__all__ = ["load_movielens100k", "load_toymat"]


def load_movielens100k():
    """
    Load the MovieLens 100k dataset fetched from: https://grouplens.org/datasets/movielens/. The version of this dataset is identical to the one include Surprise python package.

    Returns:
        pd.DataFrame: long-form dataframe user, item, rating, timestamp columns
    """
    stream = pkg_resources.resource_stream(__name__, "data/movielens100k.csv")
    return pd.read_csv(stream)


def load_toymat(users=50, items=100, random_state=None):
    """
    Generate a toy user x item dataframe

    Args:
        users (int, optional): number of users. Defaults to 50.
        items (int, optional): number of items. Defaults to 100.
        random_state ([type], optional): random state for reproducibility. Defaults to None.

    Returns:
        pd.DataFrame: users x items dataframe of ratings
    """

    random = check_random_state(random_state)
    rat = random.rand(users, items) * 50
    for x in np.arange(0, rat.shape[1], 5):
        rat[0 : int(users / 2), x] = rat[0 : int(users / 2), x] + x
    for x in np.arange(0, rat.shape[1], 3):
        rat[int(users / 2) : users, x] = rat[int(users / 2) : users, x] + x
    rat[int(users / 2) : users] = rat[int(users / 2) : users, ::-1]
    rat = pd.DataFrame(rat)
    rat.index.name = "User"
    rat.columns.name = "Item"
    return pd.DataFrame(rat)



"""
Define pytest fixtures, i.e. reusable test initializations or parameters that can be used to automatically generated a grid of tests by test functions. Brief explanation on how they work and how to write new fixtures + tests:

Each function below is passed in as a argument to a test function in one of the test_*.py files. The value of that argument == whatever the function definition returns in this file. For example:

# From test_models.py
test_init_and_dilate(init, mask, n_mask_items) <- These arguments are functions defined in *this* file.

init() <- returns an initialized model or skips a test
So within test_init_and_dilate(), init == model instance

At the same time, fixtures in this file can make use of *other* fixtures passed in as arguments, e.g.
mask(request, simulate_wide_data) <- simulate_wide_data() is defined below and returns a dataframe available inside of mask()

Arguments to fixtures that are not other fixtures, such as `request` in mask(), are special arguments that can create test grids based on an iterable of parameter values. These values are defined using the @fixture decorator. So `request` in mask() is defined with 2 parameter values: None and "masked". This means that wherever mask() is invoked in other tests, it will be called 2x: once with None and once with "masked".

This is used for example in init() where for some tests a model instance is created with masking and for other tests without masking. Then whatever tests make use of init(), such as test_init_and_dilate() will be run at least twice because mask() takes a `request` with 2 parameter values.

While a bit complicated at first, this make it easy to create testing grids based on the dependencies between fixtures. Here's a simplified example dependency graph (ignoring other fixtures):

Test 1:
simulate_wide_data() -> returns df -> mask()
request.param == None -> mask()
mask() -> init()
init() -> test_init_and_dilate() -> tests non-masked model initialization

Test 2:
simulate_wide_data() -> returns df -> mask()
request.param == "masked" -> mask()
mask() -> init()
init() -> test_init_and_dilate() -> tests masked model initialization

In reality there is an entire grid of tests because init() accepts several other fixtures which also have their own parameterizations, e.g. n_mask_items() creates 4 tests with None, 0.1, 0.5, 0.9. Combining this with the 2 parameterizations for mask() results in *8 unique tests*.

Hopefully this provides a relatively clear example of how to do exhaustive testing by defining parameter grids that are automatically created by pytest based on these definitions.
"""

import pytest
from pytest import fixture
import numpy as np
import pandas as pd
from neighbors import (
    Mean,
    KNN,
    NNMF_sgd,
    NNMF_mult,
    create_sparse_mask,
)
from string import ascii_letters

## DATA FIXTURES
@fixture(scope="module")
def simulate_wide_data():
    """Generate user x item dataframe with 50 rows and 100 columns"""

    np.random.seed(0)
    i = 100
    s = 50
    rat = np.random.rand(s, i) * 50
    for x in np.arange(0, rat.shape[1], 5):
        rat[0 : int(s / 2), x] = rat[0 : int(s / 2), x] + x
    for x in np.arange(0, rat.shape[1], 3):
        rat[int(s / 2) : s, x] = rat[int(s / 2) : s, x] + x
    rat[int(s / 2) : s] = rat[int(s / 2) : s, ::-1]
    rat = pd.DataFrame(rat)
    letters = list(ascii_letters)
    letters += [f"{elem}1" for elem in letters]
    rat.index = letters[: rat.shape[0]]
    rat.columns = letters[: rat.shape[1]]
    rat.index.name = "User"
    rat.columns.name = "Item"
    return rat


@fixture(scope="module")
def simulate_long_data(simulate_wide_data):
    """Melt generated user x item dataframe to a (user * item, 3) shaped dataframe"""

    return simulate_wide_data.reset_index().melt(id_vars="User", value_name="Rating")


@fixture(scope="module")
def simulate_simple_dataframe():
    """Simple data frame with 5 users and 2 items"""
    ratings_dict = {
        "User": ["A", "A", "B", "B", "C", "C", "D", "D", "E", "E"],
        "Item": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
        "Rating": [1, 2, 2, 4, 2.5, 4, 4.5, 5, 3, 1],
    }
    return pd.DataFrame(ratings_dict)


## INIT FIXTURES
@fixture(scope="module", params=[Mean, KNN, NNMF_mult, NNMF_sgd])
def Model(request):
    """All model classes"""
    return request.param


@fixture(scope="module", params=[None, "masked"])
def mask(request, simulate_wide_data):
    """Masked or non masked input data"""
    if request.param == "masked":
        return create_sparse_mask(simulate_wide_data, n_mask_items=0.5)
    else:
        return request.param


@fixture(scope="module", params=[None, 0.1, 0.5, 0.9])
def n_mask_items(request):
    """Percentage of items to mask on init"""
    return request.param


@fixture(scope="module")
def init(Model, mask, n_mask_items, simulate_wide_data):
    """Make grid of all model init param combinations"""
    # Check that both a mask and n_mask_items together raises an error and skip making the fixture
    if mask is not None and n_mask_items is not None:
        with pytest.raises(ValueError):
            _ = Model(
                simulate_wide_data,
                mask=mask,
                n_mask_items=n_mask_items,
            )
        pytest.skip("Ambigious init fails properly - OK")
    else:
        # Otherwise put together each init param combination
        return Model(
            simulate_wide_data,
            mask=mask,
            n_mask_items=n_mask_items,
        )


## FIT FIXTURES
@fixture(scope="module")
def model(Model, simulate_wide_data, n_mask_items):
    """Initialized model with masking already performed"""
    if n_mask_items is None:
        pytest.skip("Skip testing model with dense data and no mask - OK")
    return Model(simulate_wide_data, n_mask_items=n_mask_items)


# General
@fixture(params=[None, 1, 2])
def dilate_by_nsamples(request):
    return request.param


# # KNN only models
@fixture(params=["pearson", "cosine"])
def metric(request):
    return request.param


@fixture(params=[None, 3])
def k(request):
    return request.param


## NNMF only models
@fixture(params=[None, 10])
def n_factors(request):
    return request.param


@fixture(params=[10, 100])
def n_iterations(request):
    return request.param



"""
Core algorithms for collaborative filtering
"""

import pandas as pd
import numpy as np
from .base import Base, BaseNMF
from .utils import nanpdist
from ._fit import sgd, mult
import warnings
import matplotlib.pyplot as plt
import seaborn as sns
from numba.core.errors import NumbaPerformanceWarning

__all__ = ["Mean", "KNN", "NNMF_mult", "NNMF_sgd"]


class Mean(Base):
    """
    The Mean algorithm simply uses the mean of other users to make predictions about items. It's primarily useful as a good baseline model.
    """

    def __init__(
        self, data, mask=None, n_mask_items=None, verbose=True, random_state=None
    ):
        """
        Args:
            data (pd.DataFrame): users x items dataframe
            mask (pd.DataFrame, optional): A boolean dataframe used to split the data into 'observed' and 'missing' datasets. Defaults to None.
            n_mask_items (int/float, optional): number of items to mask out, while the rest are treated as observed; Defaults to None.
            data_range (int/float, optional): max - min of the data; Default computed from the input data. This is useful to set manually in case the input data do not span the full range of possible values
            random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. randomly mask half the data given n_mask_item = .05). Passing None will generate a new random seed. Default None.
            verbose (bool; optional): print any initialization warnings; Default True

        """
        super().__init__(
            data, mask, n_mask_items, random_state=random_state, verbose=verbose
        )
        self.mean = None

    def fit(self, dilate_by_nsamples=None, axis=0, **kwargs):

        """Fit model to train data. Simply learns item-wise mean using observed (non-missing) values.

        Args:
            dilate_ts_n_samples (int): will dilate masked samples by n_samples to leverage auto-correlation in estimating time-series data
            axis (int): dimension along which to compute mean, 0 = mean across users separately by item, 1 = mean across items separately by user; Default 0

        """

        # Call parent fit which acts as a guard for non-masked data
        super().fit()

        self.dilate_mask(n_samples=dilate_by_nsamples)
        self.mean = self.masked_data.mean(skipna=True, axis=axis)
        self._predict()
        self.is_fit = True

    def _predict(self):

        """Predict missing items using other subject's item means."""

        # Always predict mean (learned on observed values) for observed and missing values
        self.predictions = pd.concat([self.mean] * self.data.shape[0], axis=1).T
        self.predictions.index = self.data.index
        self.predictions.columns = self.data.columns


class KNN(Base):
    """
    The K-Nearest Neighbors algorithm makes predictions using a weighted mean of a subset of similar users. Similarity can be controlled via the `metric` argument to the `.fit` method, and the number of other users can be controlled with the `k` argument to the `.predict` method. NOTE: If user similiarity cannot be computed or no observed ratings have been made by the top k simililar users, this algorithm will fallback to the global mean on observed data for prediction (i.e. like the `Mean` model).
    """

    def __init__(
        self, data, mask=None, n_mask_items=None, verbose=True, random_state=None
    ):
        """
        Args:
            data (pd.DataFrame): users x items dataframe
            mask (pd.DataFrame, optional): A boolean dataframe used to split the data into 'observed' and 'missing' datasets. Defaults to None.
            n_mask_items (int/float, optional): number of items to mask out, while the rest are treated as observed; Defaults to None.
            data_range (int/float, optional): max - min of the data; Default computed from the input data. This is useful to set manually in case the input data do not span the full range of possible values
            random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. randomly mask half the data given n_mask_item = .05). Passing None will generate a new random seed. Default None.
            verbose (bool; optional): print any initialization warnings; Default True

        """
        super().__init__(
            data, mask, n_mask_items, random_state=random_state, verbose=verbose
        )
        self.user_similarity = None
        self.metric = None

    def __repr__(self):
        return f"{super().__repr__()[:-1]}, similarity_metric={self.metric})"

    def fit(
        self,
        k=10,
        metric="correlation",
        axis=0,
        dilate_by_nsamples=None,
        skip_refit=False,
        **kwargs,
    ):

        """Fit collaborative model to train data.  Calculate similarity between subjects across items. Repeated called to fit with different k, but the same previous arguments will re-use the computed user x user similarity matrix.

        Args:
            k (int): maximum number of other users to use when making a prediction for a single user. If set to None will use all users. Default 10. Note: it's possible for predictions to come from fewer than k other users if a particular user has fewer similar neighbors with positive similarity scores.
            metric (str; optional): type of similarity. One of 'correlation', 'spearman', 'kendall', 'cosine', or 'pearson'. 'pearson' is just an alias for 'correlation'. Default 'correlation'.
            axis (int): dimension along which to compute mean, 0 = mean across users separately by item, 1 = mean across items separately by user; Default 0
            skip_refit (bool; optional): skip re-estimation of user x user similarity matrix. Faster if only exploring different k and no other model parameters or masks are changing. Default False.
        """

        metrics = ["pearson", "spearman", "kendall", "cosine", "correlation"]
        if metric not in metrics:
            raise ValueError(f"metric must be one of {metrics}")

        self.metric = metric

        if metric == "correlation":
            metric = "pearson"

        # Call parent fit which acts as a guard for non-masked data
        super().fit()

        # If fit is being called more than once in a row with different k, but no other arguments are changing, reuse the last computed similarity matrix to save time. Otherwise re-calculate it
        if not skip_refit:
            self.dilate_mask(n_samples=dilate_by_nsamples)

            # Store the mean because we'll use it in cases we can't make a prediction
            self.mean = self.masked_data.mean(skipna=True, axis=axis)

            if metric in ["pearson", "kendall", "spearman"]:
                # Fall back to pandas
                sim = self.masked_data.T.corr(method=metric)
            else:
                # Convert distance metrics to similarity (currently only cosine)
                sim = pd.DataFrame(
                    1 - nanpdist(self.masked_data.to_numpy(), metric=metric),
                    index=self.masked_data.index,
                    columns=self.masked_data.index,
                )

            self.user_similarity = sim
        self._predict(k=k)
        self.is_fit = True

    def _predict(self, k):
        """Make predictions using computed user similarities.

        Args:
            k (int): number of closest neighbors to use

        """

        predictions = self.masked_data.copy()

        for row_idx, _ in self.masked_data.iterrows():

            user_prediction_error = False
            # Get the similarity of this user to all other users, ignoring self-similarity
            top_user_sims = self.user_similarity.loc[row_idx].drop(row_idx)
            if top_user_sims.isnull().all():
                warnings.warn(
                    f"User {row_idx} has no variance in their ratings. Impossible to compute similarity with other users. Falling back to global mean for all predictions"
                )
                user_prediction_error = True  # can't predict
            else:
                # Remove nan users and sort
                top_user_sims = top_user_sims[~top_user_sims.isnull()].sort_values(
                    ascending=False
                )
                if len(top_user_sims) == 0:
                    user_prediction_error = True  # can't predict
                else:
                    # Get top k if requested
                    if k is not None:
                        top_user_sims = top_user_sims[: k + 1]

                    # Rescale similarity scores to the range 0 - 1, which has the effect of zeroing out negative similarities for currently supported similarity metrics.
                    # NOTE: we should revisit this approach for non-normalized similarity metrics e.g. euclidean distance
                    top_user_sims = top_user_sims.clip(lower=0, upper=1)

                    # No top users with positive correlations
                    if len(np.nonzero(top_user_sims.to_numpy())[0]) == 0:
                        user_prediction_error = True
                    else:
                        # NOTE: this code block is just a vectorized version of looping over every item for the current user and seeing whether we have observed ratings for each of the k other users to make a prediction with. We do this because for each item the *actual* number of other users' data availble for prediction will vary between 0-k based the pattern of sparsity

                        # Get the observed ratings from top users
                        top_user_ratings = self.masked_data.loc[top_user_sims.index, :]

                        # Make predictions = user_similarity_scores (column vector) * user x item (matrix of observed ratings)
                        # Do this in pandas rather than numpy because numpy will return nans when summing items if any item is nan
                        # Yields user x item matrix of ratings scaled by similarities
                        preds = (top_user_sims * top_user_ratings.T).T
                        # Add up the ratings from other users ignoring NaNs; this serves as the numerator of the formula
                        rating_sums = preds.sum()

                        # Now some of the values in preds will be nan because we never observed a rating for that user + item combo. We need to know how many are nans and which exact ones, because we need to sum down users for preds and then divide by the sum of the similarity weights we did end up using.
                        # Get locations of where we were able to make a prediction.
                        preds_mask = ~preds.isnull()
                        # Broadcast the user similarity vector over the user x item matrix so each column now contains the user similarity score if observed a prediction from that user and a 0 if not (True is converted to 1 during this multiplication whereas False is converted to 0)
                        user_sims_mat = (preds_mask.T * top_user_sims).T
                        # Now we can just sum down the rows which will give us the sum of the similarity weights we actually used
                        sim_sums = user_sims_mat.sum()
                        # Finally get the predictions by dividing the sum of ratings by sum of similarities we ended up using. This is how Surprise does it too: https://github.com/NicolasHug/Surprise/blob/master/surprise/prediction_algorithms/knns.py#L124
                        preds = rating_sums / sim_sums

                        # For items we can't predict because we never observed any ratings from the top k users for that item, fill in with the global mean for that item
                        if preds.isnull().any():
                            preds[preds.isnull()] = self.mean[preds.isnull()]

                        predictions.loc[row_idx] = preds.to_numpy()

            # Handle cases where we were unable to make any predictions for this user
            if user_prediction_error:
                warnings.warn(
                    f"Not enough similar users with data to make any predictions for user {row_idx}. Falling back to global mean for all predictions"
                )
                predictions.loc[row_idx, :] = self.mean.to_numpy()

        self.predictions = predictions

    def plot_user_similarity(
        self, figsize=(8, 8), label_fontsize=16, hide_title=False, heatmap_kwargs={}
    ):
        """
        Plot a heatmap of user x user similarities learned on the observed data

        Args:
            figsize (tuple, optional): matplotlib figure size. Defaults to (8, 8).
            label_fontsize (int; optional): fontsize for title text; Default 16
            hide_title (bool; optional): hide title containing metric information; Default False
            heatmap_kwargs (dict, optional): addition arguments to seaborn.heatmap.

        Returns:
            ax: matplotib axis handle
        """
        if not self.is_fit:
            raise ValueError("Model as not been fit")
        if self.metric in ["correlation", "pearson", "spearman"]:
            vmin, vmax = -1, 1
            cmap = "RdBu_r"
        else:
            vmin, vmax = 0, 1
            cmap = None

        _, ax = plt.subplots(1, 1, figsize=figsize)
        _ = ax.set(xlabel=None, ylabel=None)

        ax = sns.heatmap(
            self.user_similarity,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            square=True,
            ax=ax,
            **heatmap_kwargs,
        )
        if not hide_title:
            _ = ax.set_title(f"Metric: {self.metric}", fontsize=label_fontsize)
        return ax


class NNMF_mult(BaseNMF):
    """
    The non-negative matrix factorization algorithm tries to decompose a users x items matrix into two additional matrices: users x factors and factors x items.

    Training is performed via multiplicative updating and continues until convergence or the maximum number of training iterations has been reached. Unlike the `NNMF_sgd`, this implementation takes no hyper-parameters and thus is simpler and faster to use, but less flexible, i.e. no regularization.

    The number of factors, convergence, and maximum iterations can be controlled with the `n_factors`, `tol`, and `max_iterations` arguments to the `.fit` method. By default the number of factors = the number items.

    The implementation here follows closely that of Lee & Seung, 2001 (eq 4): https://papers.nips.cc/paper/2000/file/f9d1152547c0bde01830b7e8bd60024c-Paper.pdf

    *Note*: `random_state` does not control the sgd fit, only the initialization of the factor matrices
    """

    def __init__(
        self, data, mask=None, n_mask_items=None, verbose=True, random_state=None
    ):
        """
        Args:
            data (pd.DataFrame): users x items dataframe
            mask (pd.DataFrame, optional): A boolean dataframe used to split the data into 'observed' and 'missing' datasets. Defaults to None.
            n_mask_items (int/float, optional): number of items to mask out, while the rest are treated as observed; Defaults to None.
            data_range (int/float, optional): max - min of the data; Default computed from the input data. This is useful to set manually in case the input data do not span the full range of possible values
            random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. randomly mask half the data given n_mask_item = .05). Passing None will generate a new random seed. Default None.
            verbose (bool; optional): print any initialization warnings; Default True

        """
        super().__init__(
            data, mask, n_mask_items, random_state=random_state, verbose=verbose
        )
        self.H = None  # factors x items
        self.W = None  # user x factors
        self.n_factors = None

    def __repr__(self):
        return f"{super().__repr__()[:-1]}, n_factors={self.n_factors})"

    def fit(
        self,
        n_factors=None,
        n_iterations=1000,
        tol=1e-6,
        eps=1e-6,
        verbose=False,
        dilate_by_nsamples=None,
        **kwargs,
    ):

        """Fit NNMF collaborative filtering model to train data using multiplicative updating.

        Given non-negative matrix `V` find non-negative factors `W` and `H` by minimizing `||V - WH||^2`.

        Args:
            n_factors (int, optional): number of factors to learn. Defaults to None which includes all factors.
            n_iterations (int, optional): total number of training iterations if convergence is not achieved. Defaults to 5000.
            tol (float, optional): Convergence criteria. Model is considered converged if the change in error during training < tol. Defaults to 0.001.
            eps (float; optiona): small value added to denominator of update rules to avoid divide-by-zero errors; Default 1e-6.
            verbose (bool, optional): print information about training. Defaults to False.
            dilate_by_nsamples (int, optional): How many items to dilate by prior to training. Defaults to None.
        """

        # Call parent fit which acts as a guard for non-masked data
        super().fit()

        n_users, n_items = self.data.shape

        if (
            isinstance(n_factors, int) and (n_factors > n_items and n_factors > n_users)
        ) or isinstance(n_factors, np.floating):
            raise TypeError("n_factors must be an integer < number of items and users")

        if n_factors is None:
            n_factors = min([n_users, n_items])

        self.n_factors = n_factors

        # Initialize W and H as non-negative scaled random values
        # We use random initialization scaled by the number of factors not unlike sklearn: https://github.com/scikit-learn/scikit-learn/blob/95119c13af77c76e150b753485c662b7c52a41a2/sklearn/decomposition/_nmf.py#L334
        self.W = np.abs(
            self.random_state.normal(scale=1.0 / n_factors, size=(n_users, n_factors))
        )
        self.H = np.abs(
            self.random_state.normal(scale=1.0 / n_factors, size=(n_factors, n_items))
        )

        # Whereas in SGD we explity pass in indices of training data for fitting, here we set testing indices to 0 so they have no impact on the multiplicative update. See Zhu, 2016 for more details: https://arxiv.org/pdf/1612.06037.pdf
        self.dilate_mask(n_samples=dilate_by_nsamples)

        # fillna(0) is equivalent to hadamard (element-wise) product with a binary mask
        X = self.masked_data.fillna(0).to_numpy()

        # Run multiplicative updating
        # Silence numba warning until this issue gets fixed: https://github.com/numba/numba/issues/4585
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=NumbaPerformanceWarning)
            error_history, converged, n_iter, delta, norm_rmse, W, H = mult(
                X,
                self.W,
                self.H,
                self.data_range,
                eps,
                tol,
                n_iterations,
                verbose,
            )

        # Save outputs to model
        self.W, self.H = W, H
        self.error_history = error_history
        self._n_iter = n_iter
        self._delta = delta
        self._norm_rmse = norm_rmse
        self.converged = converged

        if verbose:
            if self.converged:
                print("\n\tCONVERGED!")
                print(f"\n\tFinal Iteration: {self._n_iter}")
                print(f"\tFinal Delta: {np.round(self._delta)}")
            else:
                print("\tFAILED TO CONVERGE (n_iter reached)")
                print(f"\n\tFinal Iteration: {self._n_iter}")
                print(f"\tFinal delta exceeds tol: {tol} <= {self._delta}")

            print(f"\tFinal Norm Error: {np.round(100*norm_rmse, 2)}%")
        self._predict()
        self.is_fit = True

    def _predict(self):

        """Predict subjects' missing items using NNMF with multiplicative updating"""

        self.predictions = pd.DataFrame(
            self.W @ self.H, index=self.data.index, columns=self.data.columns
        )


class NNMF_sgd(BaseNMF):
    """
    The non-negative matrix factorization algorithm tries to decompose a users x items matrix into two additional matrices: users x factors and factors x items.

    Training is performed via stochastic-gradient-descent and continues until convergence or the maximum number of iterations has been reached. Unlike `NNMF_mult` errors during training are used to update latent factors *separately* for each user/item combination. Additionally this implementation is more flexible as it supports hyperparameters for various kinds of regularization at the cost of increased computation time.

    The number of factors, convergence, and maximum iterations can be controlled with the `n_factors`, `tol`, and `max_iterations` arguments to the `.fit` method. By default the number of factors = the number items.

    `random_state` does not control the sgd fit, only the initialization of the factor matrices

    **Important Note**: model fitting can be highly sensitive to the regularization hyper-parameters passed to `.fit`. These hyper-parameters control the amount of regularization used when learning user and item factors and biases. By default *no regularization* is performed. For some combinations of hyper-parameters (e.g. large `user_fact_reg` and small `item_fact_reg`) latent vectors can blow up to infinity producing `NaNs` in model estimates. Model fitting will not fail in these cases so **caution** should be taken when making use of hyper-parameters.

    """

    def __init__(
        self, data, mask=None, n_mask_items=None, verbose=True, random_state=None
    ):
        """
        Args:
            data (pd.DataFrame): users x items dataframe
            mask (pd.DataFrame, optional): A boolean dataframe used to split the data into 'observed' and 'missing' datasets. Defaults to None.
            n_mask_items (int/float, optional): number of items to mask out, while the rest are treated as observed; Defaults to None.
            data_range (int/float, optional): max - min of the data; Default computed from the input data. This is useful to set manually in case the input data do not span the full range of possible values
            random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. randomly mask half the data given n_mask_item = .05). Passing None will generate a new random seed. Default None.
            verbose (bool; optional): print any initialization warnings; Default True

        """
        super().__init__(
            data, mask, n_mask_items, random_state=random_state, verbose=verbose
        )
        self.n_factors = None

    def __repr__(self):
        return f"{super().__repr__()[:-1]}, n_factors={self.n_factors})"

    def fit(
        self,
        n_factors=None,
        item_fact_reg=0.0,
        user_fact_reg=0.0,
        item_bias_reg=0.0,
        user_bias_reg=0.0,
        learning_rate=0.001,
        n_iterations=1000,
        tol=1e-6,
        verbose=False,
        dilate_by_nsamples=None,
        **kwargs,
    ):
        """
        Fit NNMF collaborative filtering model using stochastic-gradient-descent. **Note:** Some combinations of fit parameters may lead to degenerate fits due to use and item vectors converging to infinity. Because no constraints are imposed on the values these parameters can take, please adjust them with caution. If you encounter NaNs in your predictions it's likely because of the specific combination of parameters you chose and you can try refitting with the default settings (i.e. no regularization and learning rate = 0.001). Use `verbose=True` to help determine at what iteration these degenerate fits occur.

        Args:
            n_factors (int, optional): number of factors to learn. Defaults to None which includes all factors.
            item_fact_reg (float, optional): item factor regularization to apply. Defaults to 0.0.
            user_fact_reg (float, optional): user factor regularization to apply. Defaults to 0.0.
            item_bias_reg (float, optional): item factor bias term to apply. Defaults to 0.0.
            user_bias_reg (float, optional): user factor bias term to apply. Defaults to 0.0.
            learning_rate (float, optional): how quickly to integrate errors during training. Defaults to 0.001.
            n_iterations (int, optional): total number of training iterations if convergence is not achieved. Defaults to 5000.
            tol (float, optional): Convergence criteria. Model is considered converged if the change in error during training < tol. Defaults to 0.001.
            verbose (bool, optional): print information about training. Defaults to False.
            dilate_by_nsamples (int, optional): How many items to dilate by prior to training. Defaults to None.
        """

        # Call parent fit which acts as a guard for non-masked data
        super().fit()

        # initialize variables
        n_users, n_items = self.data.shape

        if (
            isinstance(n_factors, int) and (n_factors > n_items and n_factors > n_users)
        ) or isinstance(n_factors, np.floating):
            raise TypeError("n_factors must be an integer < number of items and users")

        if n_factors is None:
            n_factors = min([n_users, n_items])

        self.n_factors = n_factors
        self.item_fact_reg = item_fact_reg
        self.user_fact_reg = user_fact_reg
        self.item_bias_reg = item_bias_reg
        self.user_bias_reg = user_bias_reg
        self.error_history = []

        # Perform dilation if requested
        self.dilate_mask(n_samples=dilate_by_nsamples)

        # Get indices of training data to compute; np.nonzero returns a tuple of row and column indices that when iterated over simultaneosly yield the [row_index, col_index] of each training observation
        if self.is_mask_dilated:
            row_indices, col_indices = self.dilated_mask.values.nonzero()
        else:
            row_indices, col_indices = self.mask.values.nonzero()

        # Convert tuples cause numba complains
        row_indices, col_indices = np.array(row_indices), np.array(col_indices)

        # Initialize global, user, and item biases and latent vectors
        self.global_bias = self.masked_data.mean().mean()
        self.user_bias = np.zeros(n_users)
        self.item_bias = np.zeros(n_items)

        # Initialize random values oriented these as user x factor, factor x item
        self.user_vecs = np.abs(
            self.random_state.normal(scale=1.0 / n_factors, size=(n_users, n_factors))
        )
        self.item_vecs = np.abs(
            self.random_state.normal(scale=1.0 / n_factors, size=(n_factors, n_items))
        )

        X = self.masked_data.to_numpy()

        # Generate seed for shuffling within sgd
        seed = self.random_state.randint(np.iinfo(np.int32).max)

        # Run SGD
        # Silence numba warning until this issue gets fixed: https://github.com/numba/numba/issues/4585
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=NumbaPerformanceWarning)
            (
                error_history,
                converged,
                n_iter,
                delta,
                norm_rmse,
                user_bias,
                user_vecs,
                item_bias,
                item_vecs,
            ) = sgd(
                X,
                seed,
                self.global_bias,
                self.data_range,
                tol,
                self.user_bias,
                self.user_vecs,
                self.user_bias_reg,
                self.user_fact_reg,
                self.item_bias,
                self.item_vecs,
                self.item_bias_reg,
                self.item_fact_reg,
                n_iterations,
                row_indices,
                col_indices,
                learning_rate,
                verbose,
            )
        # Save outputs to model
        (
            self.error_history,
            self.user_bias,
            self.user_vecs,
            self.item_bias,
            self.item_vecs,
        ) = (
            error_history,
            user_bias,
            user_vecs,
            item_bias,
            item_vecs,
        )

        self._n_iter = n_iter
        self._delta = delta
        self._norm_rmse = norm_rmse
        self.converged = converged
        if verbose:
            if self.converged:
                print("\n\tCONVERGED!")
                print(f"\n\tFinal Iteration: {self._n_iter}")
                print(f"\tFinal Delta: {np.round(self._delta)}")
            else:
                print("\tFAILED TO CONVERGE (n_iter reached)")
                print(f"\n\tFinal Iteration: {self._n_iter}")
                print(f"\tFinal delta exceeds tol: {tol} <= {self._delta}")

            print(f"\tFinal Norm Error: {np.round(100*norm_rmse, 2)}%")

        self._predict()
        self.is_fit = True

    def _predict(self):

        """Predict User's missing items using NNMF with stochastic gradient descent"""

        # user x factor * factor item + biases
        predictions = self.user_vecs @ self.item_vecs
        predictions = (
            (predictions.T + self.user_bias).T + self.item_bias + self.global_bias
        )
        self.predictions = pd.DataFrame(
            predictions, index=self.data.index, columns=self.data.columns
        )



# flake8: noqa
from .base import Base, BaseNMF
from .models import Mean, KNN, NNMF_mult, NNMF_sgd
from .utils import (
    create_user_item_matrix,
    invert_user_item_matrix,
    get_size_in_mb,
    get_sparsity,
    nanpdist,
    create_sparse_mask,
    estimate_performance,
    flatten_dataframe,
    unflatten_dataframe,
    split_train_test,
    check_random_state,
)
from .data import load_movielens100k, load_toymat
from .version import __version__



"""
Utility functions and helpers
"""

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.spatial.distance import pdist, squareform
import numbers
from itertools import product, chain
from typing import Union
from concurrent.futures import ThreadPoolExecutor
import os
import time

__all__ = [
    "create_user_item_matrix",
    "invert_user_item_matrix",
    "get_size_in_mb",
    "get_sparsity",
    "nanpdist",
    "create_sparse_mask",
    "estimate_performance",
    "flatten_dataframe",
    "unflatten_dataframe",
    "split_train_test",
    "check_random_state",
]


def check_random_state(seed):
    """
    Turn seed into a np.random.RandomState instance, lifted directly from sklearn to avoid a heavy dependency for one function.

    Olivier Grisel, Andreas Mueller, Lars, Alexandre Gramfort, Gilles Louppe, Peter Prettenhofer, … Eustache. (2021, January 19). scikit-learn/scikit-learn: scikit-learn 0.24.1 (Version 0.24.1). Zenodo. http://doi.org/10.5281/zenodo.4450597

    Args:
        seed (None, int, RandomState): If seed is None, return the RandomState singleton used by np.random. If seed is an int, return a new RandomState instance seeded with seed. If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """

    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        "%r cannot be used to seed a numpy.random.RandomState" " instance" % seed
    )


def invert_user_item_matrix(df):
    """
    Reshapes a user x item matrix back into a longform dataframe with "User", "Item" and "Rating" columns. Convenience function to undo a call to `create_user_item_matrix`. Dataframe must have an index name of 'Users' and a columns name of 'Items'.

    Args:
        df (pd.DataFrame): user x item dataframe, e.g. the output of a call to `create_user_by_item_matrix`

    Returns:
        pd.DataFrame: longform dataframe with "User", "Item", "Rating" columns
    """

    if df.index.name != "User" or df.columns.name != "Item":
        raise TypeError(
            "input should be a user x item matrix with appropriately named columns and rows"
        )
    return df.reset_index().melt(id_vars="User", value_name="Rating")


def create_user_item_matrix(df, columns=None, force_float=True, errors="raise"):

    """Convert a longform dataframe containing columns with unique user ids, item ids, and ratings into a user x item wide matrix

    Args:
        df (Dataframe): input dataframe
        columns (list): list of length 3 with dataframe columns to use for reshaping. The first value should reflect unique individual identifier ("User"), the second a unique item identifier (e.g. "Item"), and the last the rating made by the individual on that item ("Rating"). Defaults to ["User", "Item", "Rating"]
        force_float (bool): force the resulting output to be float data types with errors being set to NaN; Default True
        errors (string): how to handle errors in pd.to_numeric; Default 'raise'

    Return:
        pd.DataFrame: user x item rating Dataframe

    """

    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be pandas instance")
    if columns is None:
        columns = ["User", "Item", "Rating"]
    if not all([x in df.columns for x in columns]):
        raise ValueError(
            f"df is missing some or all of the following columns: {columns}"
        )

    ratings = df[columns]
    ratings = ratings.pivot(index=columns[0], columns=columns[1], values=columns[2])
    try:
        if force_float:
            ratings = ratings.apply(pd.to_numeric, errors=errors)
    except ValueError as e:
        print(
            "Auto-converting data to floats failed, probably because you have non-numeric data in some rows. You can set errors = 'coerce' to set these failures to NaN"
        )
        raise (e)
    ratings.index.name = "User"
    ratings.columns.name = "Item"

    return ratings


def get_size_in_mb(arr):
    """Calculates size of ndarray in megabytes"""
    if isinstance(arr, (np.ndarray, csr_matrix)):
        return arr.data.nbytes / 1e6
    else:
        raise TypeError("input must by a numpy array or scipy csr sparse matrix")


def get_sparsity(arr):
    """Calculates sparsity of ndarray (0 - 1)"""
    if isinstance(arr, np.ndarray):
        return 1 - (np.count_nonzero(arr) / arr.size)
    elif isinstance(arr, pd.DataFrame):
        return arr.isnull().sum().sum() / arr.size
    else:
        raise TypeError("input must be a numpy array")


# Can try to speed this up with numba, but lose support for pandas and scipy so we'd have to rewrite distance functions in numpy/python
def nanpdist(arr, metric="euclidean", return_square=True):
    """
    Just like scipy.spatial.distance.pdist or sklearn.metrics.pairwise_distances, but respects NaNs by only comparing the overlapping values from pairs of rows.

    Args:
        arr (np.ndarray): 2d array
        metric (str; optional): distance metric to use. Must be supported by scipy
        return_square (boo; optional): return a symmetric 2d distance matrix like sklearn instead of a 1d vector like pdist; Default True

    Return:
        np.ndarray: symmetric 2d array of distances
    """

    has_nans = pd.DataFrame(arr).isnull().any().any()
    if not has_nans:
        out = pdist(arr, metric=metric)
    else:
        nrows = arr.shape[0]
        out = np.zeros(nrows * (nrows - 1) // 2, dtype=float)
        mask = np.isfinite(arr)
        vec_mask = np.zeros(arr.shape[1], dtype=bool)
        k = 0
        for row1_idx in range(nrows - 1):
            for row2_idx in range(row1_idx + 1, nrows):
                vec_mask = np.logical_and(mask[row1_idx], mask[row2_idx])
                masked_row1, masked_row2 = (
                    arr[row1_idx][vec_mask],
                    arr[row2_idx][vec_mask],
                )
                out[k] = pdist(np.vstack([masked_row1, masked_row2]), metric=metric)
                k += 1

    if return_square:
        if out.ndim == 1:
            out = squareform(out)
    else:
        if out.ndim == 2:
            out = squareform(out)
    return out


def create_sparse_mask(data, n_mask_items=0.2, random_state=None):
    """
    Given a pandas dataframe create a boolean mask such that n_mask_items columns are `False` and the rest are `True`. Critically, each row is masked independently. This function does not alter the input dataframe.

    Args:
        data (pd.DataFrame): input dataframe
        n_train_items (float, optional): if an integer is passed its raw value is used. Otherwise if a float is passed its taken to be a (rounded) percentage of the total items. Defaults to 0.1 (10% of the columns of data).

    Raises:
        TypeError: [description]

    Returns:
        pd.DataFrame: boolean dataframe of same shape as data
    """

    if data.isnull().any().any():
        raise ValueError("data already contains NaNs and further masking is ambiguous!")

    if isinstance(n_mask_items, (float, np.floating)) and 1 >= n_mask_items > 0:
        n_false_items = int(np.round(data.shape[1] * n_mask_items))

    elif isinstance(n_mask_items, (int, np.integer)):
        n_false_items = n_mask_items

    else:
        raise TypeError(
            f"n_train_items must be an integer or a float between 0-1, not {type(n_mask_items)} with value {n_mask_items}"
        )

    random = check_random_state(random_state)
    n_true_items = data.shape[1] - n_false_items
    mask = np.array([True] * n_true_items + [False] * n_false_items)
    mask = np.vstack(
        [
            random.choice(mask, replace=False, size=mask.shape)
            for _ in range(data.shape[0])
        ]
    )
    return pd.DataFrame(mask, index=data.index, columns=data.columns)


def flatten_dataframe(data: pd.DataFrame) -> list:
    """
    Given a 2d dataframe return a numpy array of arrays organized as (row_idx, col_idx, val). This function is analgous to numpy.ravel or numpy.flatten for arrays, with the addition of the row and column indices for each value

    Args:
        data (pd.DataFrame): input dataframe

    Returns:
        np.ndarray: arrayo of row, column, value triplets
    """

    if not isinstance(data, pd.DataFrame):
        raise TypeError("input must be a pandas dataframe")

    out = zip(product(data.index, data.columns), data.to_numpy().ravel())
    return np.array([(elem[0][0], elem[0][1], elem[1]) for elem in out])


def unflatten_dataframe(
    data: np.ndarray,
    columns=None,
    index=None,
    num_rows=None,
    num_cols=None,
    index_name=None,
    columns_name=None,
) -> pd.DataFrame:
    """
    Reverse a flatten_dataframe operation to reconstruct the original unflattened dataframe

    Args:
        data (np.ndarray): n_items x 3 numpy array where columns represent row_idx, col_idx, and val at the location.
        columns (list, optional): column names of new dataframe. Defaults to None.
        index (list, optional): row names of new dataframe. Defaults to None.
        num_rows (int, optional): total number of rows. Useful if the flattened dataframe had a non-numerical non-ordered index. Default None which uses the max(row_idx)
        num_cols (int, optional): total number of cols. Useful if the flattened dataframe had a non-numerical non-ordered index. Default None which uses the max(col_idx)
        index_name (str; optional): Name of rows; Default None
        columns_name (str; optional): Name of columns; Default None

    Returns:
        pd.DataFrame: original unflattened dataframe
    """

    if not isinstance(data, np.ndarray):
        raise TypeError("input should be a numpy array")
    if index is None and num_rows is None:
        index = list(dict.fromkeys(data[:, 0]))
        num_rows = len(index)
    elif index is not None and num_rows is None:
        num_rows = len(index)
    elif index is None and num_rows is not None:
        index = list(dict.fromkeys(data[:, 0]))
        if len(index) != num_rows:
            raise ValueError(
                "num_rows does not match the number of unique row_idx values in data"
            )
    if columns is None and num_cols is None:
        columns = list(dict.fromkeys(data[:, 1]))
        num_cols = len(columns)
    elif columns is not None and num_cols is None:
        num_cols = len(columns)
    elif columns is None and num_cols is not None:
        columns = list(dict.fromkeys(data[:, 1]))
        if len(columns) != num_cols:
            raise ValueError(
                "num_cols does not match the number of unique col_idx values in data"
            )
    out = np.empty((num_rows, num_cols))
    out[:] = np.nan
    out = pd.DataFrame(out, index=index, columns=columns)
    for elem in data:
        out.loc[elem[0], elem[1]] = np.float(elem[2])
    out.index.name = index_name
    out.columns.name = columns_name
    return out


def downsample_dataframe(data, n_samples, sampling_freq=None, target_type="samples"):
    """
    Down sample a dataframe

    Args:
        data (pd.DataFrame): input data
        n_samples (int): number of samples.
        sampling_freq (int/float, optional): sampling frequency of data in hz. Defaults to None.
        target_type (str, optional): how to downsample; must be one of "samples", "seconds" or "hz". Defaults to "samples".

    Returns:
        pd.DataFrame: downsampled input data
    """

    if not isinstance(data, pd.DataFrame):
        raise TypeError("data must be a pandas dataframe")
    if not isinstance(n_samples, int):
        raise TypeError("n_samples must be an integer")
    if target_type not in ["samples", "seconds", "hz"]:
        raise ValueError("target_type must be 'samples', 'seconds', or 'hz'")

    if target_type in ["seconds", "hz"] and (
        n_samples is None or sampling_freq is None
    ):
        raise ValueError(
            "if target_type = {target_type}, both sampling_freq and target must be provided"
        )

    if target_type == "seconds":
        n_samples = n_samples * sampling_freq
    elif target_type == "hz":
        n_samples = sampling_freq / n_samples
    else:
        n_samples = n_samples

    data = data.T
    idx = np.sort(np.repeat(np.arange(1, data.shape[0] / n_samples, 1), n_samples))
    if data.shape[0] > len(idx):
        idx = np.concatenate([idx, np.repeat(idx[-1] + 1, data.shape[0] - len(idx))])
    return data.groupby(idx).mean().T


def split_train_test(
    data: pd.DataFrame, n_folds: int = 10, shuffle=True, random_state=None
):
    """
    Custom train/test split generator for leave-one-fold out cross-validation. Given a user x item dataframe of dense or sparse data, generates n_folds worth of train/test split dataframes that have the same shape as data, but with non-overlapping values. This ensures that no user-item combination appears in both train and test splits. Useful for estimating out-of-sample performance. n_folds controls the train/test split ratio, e.g. n_folds = 5 means train = 4/5 (~80%); test = 1/5 (~20%)

    Args:
        data (pd.DataFrame): user x item dataframe
        n_folds (int, optional): number of train/test splits to generate. Defaults to 5.
        shuffle (bool, optional): randomize what user-item combinations appear in each split. If shuffle is True, folds will split in order of item user-item appearance; Default True
        random_state (int, None, numpy.random.mtrand.RandomState): random state for reproducibility; Default None

    Yields:
        (tuple): train, test dataframes for each fold with same shape as data
    """

    random_state = check_random_state(random_state)
    num_rows, num_cols = data.shape
    flat = flatten_dataframe(data)
    if shuffle:
        random_state.shuffle(flat)
    start, stop = 0, 0
    for fold in range(n_folds):
        start = stop
        stop += len(flat) // n_folds
        if fold < len(flat) % n_folds:
            stop += 1

        train = np.array([elem for elem in chain(flat[:start], flat[stop:])])
        test = np.array([elem for elem in flat[start:stop]])

        yield unflatten_dataframe(
            train,
            num_rows=num_rows,
            num_cols=num_cols,
            index=data.index,
            columns=data.columns,
            index_name=data.index.name,
            columns_name=data.columns.name,
        ), unflatten_dataframe(
            test,
            num_rows=num_rows,
            num_cols=num_cols,
            index=data.index,
            columns=data.columns,
            index_name=data.index.name,
            columns_name=data.columns.name,
        )


def estimate_performance(
    algorithm,
    data: pd.DataFrame,
    n_iter: int = 10,
    n_folds: int = 10,
    n_mask_items: Union[int, np.floating] = 0.2,
    return_agg: bool = True,
    return_full_performance: bool = False,
    agg_stats: tuple = ["mean", "std"],
    fit_kwargs: dict = {},
    random_state=None,
    parallelize=False,
    timeit=True,
    verbose=False,
) -> pd.DataFrame:
    """
    Repeatedly fit a model with data to estimate performance. If input data is dense (contains no missing values) then this function will fit the model `n_iter` times after applying a different random mask each iteration according to `n_mask_items`. This is useful for testing how an algorithm *would have performed* given data of a specified sparsity.

    On the other hand, if input data is already sparse, no further masking will be performed and the data will instead be split into `n_folds` training and testing folds. Evaluation will be performed based on how well non-missing values in testing folds are recovered. **Note**: this *increases the sparsity* of the input dataset. The number of folds requested controls the additional sparsity of each train and test split. For example, with the default `n_folds=10`, each training split will contain 9/10 folds = ~90% of *observed values* (additional sparsity of 10%); each test split will contain 1/10 folds (~10% of *observed values*).

    Args:
        algorithm (neighbors.model): an uninitialized model, e.g. `Mean`
        data (pd.DataFrame): a users x item dataframe
        n_iter (int, optional): number of repetitions for dense data. Defaults to 10.
        n_folds (int, optional): number of folds for CV on sparse data. Defaults to 10.
        n_mask_items (int/float, optional): how much randomly sparsify dense data each iteration; Defaults to masking out 20% of observed values
        return_agg (bool, optional): Return mean and std over repetitions rather than the reptitions themselves Defaults to True.
        return_full_performance (bool, optional): return the performance against both "observed" and "missing" or just "missing" values if using dense data and `n_iter`. Likewise return performance of both "train" and "test" or just "test" splits if using sparse data and `n_folds`; Default False
        agg_stats (list): string names of statistics to compute over repetitions. Must be accepted by `pd.DataFrame.agg`; Default ('mean', 'std')
        fit_kwargs (dict, optional): A dictionary of arguments passed to the `.fit()` method of the model. Defaults to {}.
        random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. random masking, cv splitting, etc). Passing None will generate a new random seed; Default None.
        parallelize (bool, optional): Use multiple *threads* to run n_iter or n_folds in parallel. To save memory and prevent data duplication, this does not use multiple *processes* like joblib. Some algorithms like `NNMF_sgd` can see significant speed-ups (2-3x) when `True`. Others, like `NNMF_mult` do not not gain much benefit or may even slow down. Default False.
        timeit (bool, option): include a column of estimation + prediction duration for each iteration/fold in the results. Ignored when return_agg=True; Default True.
        verbose (bool, optional): print information messages on execution; Default False.


    Returns:
        pd.DataFrame: aggregated or non-aggregated summary statistics
    """

    sparsity = get_sparsity(data)
    random_state = check_random_state(random_state)
    # Set max threads to the new default in Python 3.8
    max_threads = min(32, os.cpu_count() + 4)

    if sparsity == 0.0:
        # DENSE DATA so re-mask each iteration
        print(f"Data sparsity is {np.round(sparsity*100,2)}%. Using random masking...")

        seeds = random_state.randint(np.iinfo(np.int32).max, size=n_iter)

        def _run_dense(
            i,
            algorithm,
            data,
            random_state,
            n_mask_items,
            return_full_performance,
            timeit,
            fit_kwargs,
        ):
            if timeit:
                start = time.time()
            model = algorithm(
                data=data,
                random_state=random_state,
                n_mask_items=n_mask_items,
                verbose=False,
            )
            model.fit(**fit_kwargs)
            # observed + missing
            if return_full_performance:
                results = model.summary(dataset=["missing", "observed"])
            # only missing
            else:
                results = model.summary(dataset="missing")
            results["iter"] = i + 1

            # Individual user results
            user_results = model.user_results
            user_results["iter"] = i + 1

            if timeit:
                stop = time.time()
                duration = stop - start
                results["duration"] = duration
            return results, user_results

        if parallelize:
            if verbose:
                print("Parallelizing across threads...")

            with ThreadPoolExecutor(max_threads) as threads:
                all_results = threads.map(
                    _run_dense,
                    range(n_iter),
                    [algorithm] * n_iter,
                    [data] * n_iter,
                    seeds,
                    [n_mask_items] * n_iter,
                    [return_full_performance] * n_iter,
                    [timeit] * n_iter,
                    [fit_kwargs] * n_iter,
                )
        else:
            all_results = map(
                _run_dense,
                range(n_iter),
                [algorithm] * n_iter,
                [data] * n_iter,
                seeds,
                [n_mask_items] * n_iter,
                [return_full_performance] * n_iter,
                [timeit] * n_iter,
                [fit_kwargs] * n_iter,
            )

        col_order = ["algorithm", "dataset", "iter", "group", "metric", "score"]
        user_agg_drop_col = ["iter"]
    else:
        # SPARSE DATA so split observed values according to n_folds
        print(
            f"Data sparsity is {np.round(sparsity*100,2)}%. Using cross-validation..."
        )

        def _run_sparse(
            i,
            train,
            test,
            algorithm,
            random_state,
            return_full_performance,
            timeit,
            fit_kwargs,
        ):
            if timeit:
                start = time.time()
            model = algorithm(data=train, random_state=random_state, verbose=False)
            model.fit(**fit_kwargs)
            test_results = model.summary(actual=test, dataset="full")
            test_results["cv_fold"] = i + 1
            test_results["dataset"] = test_results.dataset.map({"full": "test"})

            user_test_results = model.user_results
            user_test_results["cv_fold"] = i + 1
            user_test_results.columns = [
                "rmse_test",
                "mse_test",
                "mae_test",
                "correlation_test",
                "cv_fold",
            ]

            if return_full_performance:
                train_results = model.summary(dataset="full")
                train_results["cv_fold"] = i + 1
                train_results["dataset"] = train_results.dataset.map({"full": "train"})
                test_results = pd.concat(
                    [test_results, train_results], ignore_index=True
                )

                user_train_results = model.user_results
                user_train_results.columns = [
                    "rmse_train",
                    "mse_train",
                    "mae_train",
                    "correlation_train",
                ]
                user_test_results = pd.concat(
                    [user_train_results, user_test_results], axis=1
                )

            if timeit:
                stop = time.time()
                duration = stop - start
                test_results["duration"] = duration
            return test_results, user_test_results

        seeds = random_state.randint(np.iinfo(np.int32).max, size=n_folds)
        splits = list(split_train_test(data, n_folds, random_state=random_state))
        trains = [elem[0] for elem in splits]
        tests = [elem[1] for elem in splits]

        if parallelize:
            if verbose:
                print("Parallelizing across threads...")

            with ThreadPoolExecutor(max_threads) as threads:
                all_results = threads.map(
                    _run_sparse,
                    range(n_folds),
                    trains,
                    tests,
                    [algorithm] * n_folds,
                    seeds,
                    [return_full_performance] * n_folds,
                    [timeit] * n_folds,
                    [fit_kwargs] * n_folds,
                )
        else:
            all_results = map(
                _run_sparse,
                range(n_folds),
                trains,
                tests,
                [algorithm] * n_folds,
                seeds,
                [return_full_performance] * n_folds,
                [timeit] * n_folds,
                [fit_kwargs] * n_folds,
            )

        col_order = ["algorithm", "dataset", "cv_fold", "group", "metric", "score"]
        user_agg_drop_col = ["cv_fold"]

    # Collect results
    all_results = list(all_results)
    group_results = [elem[0] for elem in all_results]
    user_results = [
        elem[1].reset_index().rename(columns={"User": "user"}) for elem in all_results
    ]
    group_results = (
        pd.concat(group_results, ignore_index=True)
        .sort_values(by=["group", "dataset", "metric"])
        .reset_index(drop=True)
    )
    user_results = pd.concat(user_results, ignore_index=True)

    # Handle aggregation
    if return_agg:
        col_order = ["algorithm", "dataset", "group", "metric"] + agg_stats
        group_results = (
            group_results.groupby(["algorithm", "dataset", "group", "metric"])
            .score.agg(agg_stats)
            .reset_index()
            .sort_values(by=["dataset", "group", "metric"])
            .reset_index(drop=True)[col_order]
        )
        user_results = (
            user_results.groupby("user").mean().drop(columns=user_agg_drop_col)
        )
        return group_results, user_results
    else:
        if timeit:
            col_order += ["duration"]
        return group_results[col_order], user_results



import pandas as pd
import numpy as np
from scipy.stats import pearsonr
from copy import deepcopy
import matplotlib.pyplot as plt
from .utils import create_sparse_mask, downsample_dataframe, check_random_state
import warnings
import seaborn as sns

__all__ = ["Base", "BaseNMF"]


class Base(object):
    """
    This is the base class for all model types.
    """

    def __init__(
        self,
        data,
        mask=None,
        n_mask_items=None,
        data_range=None,
        random_state=None,
        verbose=True,
    ):
        """
        Initialize a base collaborative filtering model

        Args:
            data (pd.DataFrame): users x items dataframe
            mask (pd.DataFrame, optional): A boolean dataframe used to split the data into 'observed' and 'missing' datasets. Defaults to None.
            n_mask_items (int/float, optional): number of items to mask out, while the rest are treated as observed; Defaults to None.
            data_range (int/float, optional): max - min of the data; Default computed from the input data. This is useful to set manually in case the input data do not span the full range of possible values
            random_state (None, int, RandomState): a seed or random state used for all internal random operations (e.g. randomly mask half the data given n_mask_item = .05). Passing None will generate a new random seed. Default None.
            verbose (bool; optional): print any initialization warnings; Default True

        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError("data must be a pandas dataframe instance")
        self.data = data.copy()
        self.predictions = None
        self.is_fit = False
        self.mask = None  # boolean matrix of observed value indices
        self.is_masked = False
        self.masked_data = data  # boolean mask applied to data
        self.is_mask_dilated = False
        self.dilated_mask = None  # booleanized masked_data after dilation
        self.dilated_by_nsamples = None
        self.n_mask_items = n_mask_items
        self.data_range = (
            self.data.max().max() - self.data.min().min()
            if data_range is None
            else data_range
        )
        self.is_dense = True
        self.overall_results = None
        self.user_results = None
        self.n_users = data.shape[0]
        self.n_items = data.shape[1]
        self.random_state = check_random_state(random_state)

        # Check for null values in input data and if they exist treat the data as already masked
        if self.data.isnull().any().any():
            if verbose:
                print("data contains NaNs...treating as pre-masked")
            self.mask = ~self.data.isnull()
            self.is_masked = True
            self.is_dense = False

        # Otherwise apply any user provided mask or tell them data was pre-masked
        if mask is not None:
            if self.is_masked:
                raise ValueError(
                    "mask was provided, but data already contains missing values that were used for masking. This is an ambiguous operation. If a mask is provided data should not contain missing values!"
                )
            if mask.shape != data.shape:
                raise ValueError("mask must be the same shape as data")
            self.mask = mask.copy()
            self.masked_data = self.data[self.mask]
            self.is_masked = True

        # Same for n_train_items
        if self.n_mask_items is not None:
            if self.is_masked:
                raise ValueError(
                    "n_mask_items was provided, but data already contains missing values that were used for masking. This is an ambiguous operation. If a n_mask_items is provided data should not contain missing values!"
                )
            self.create_masked_data(n_mask_items=self.n_mask_items)

        if mask is None and n_mask_items is None and not self.is_masked and verbose:
            print(
                "Model initialized with dense data. Make sure to utilize `.create_masked_data` prior to fitting or reinitialize with a mask or n_mask_items"
            )

    def __repr__(self):
        out = f"{self.__class__.__module__}.{self.__class__.__name__}(n_users={self.n_users}, n_items={self.n_items}, is_fit={self.is_fit}, is_masked={self.is_masked}, is_mask_dilated={self.is_mask_dilated}"
        if self.is_mask_dilated:
            out += f", dilated_by_nsamples={self.dilated_by_nsamples}"
        out += ")"
        return out

    def score(
        self,
        metric="rmse",
        dataset="missing",
        by_user=True,
        actual=None,
    ):
        """Get the performance of a fitted model by comparing observed and predicted data. This method is primarily useful if you want to calculate a single metric. Otherwise you should prefer the `.summary()` method instead, which scores all metrics.

        Args:
            metric (str; optional): what metric to compute, one of 'rmse', 'mse', 'mae' or 'correlation'; Default 'rmse'.
            dataset (str; optional): how to compute scoring, either using 'observed', 'missing' or 'full'. Default 'missing'.
            by_user (bool; optional): whether to return a single score over all data points or a pandas Series of scores per user. Default True.
            actual (pd.DataFrame, None; optional): a dataframe to score against; Default is None which uses the data provided when the model was initialized

        Returns:
            float/pd.Series: score

        """

        if not self.is_fit:
            raise ValueError("You must fit() model first before using this method.")

        if metric not in ["rmse", "mse", "mae", "correlation"]:
            raise ValueError(
                "metric must be one of 'rmse', 'mse', 'mae', or 'correlation'"
            )
        # Get dataframes of observed and predicted values
        # This will be a dense or sparse matrix the same shape as the input data
        model_actual, pred = self._retrieve_predictions(dataset)

        if actual is None:
            actual = model_actual
        else:
            if actual.shape != self.data.shape:
                raise ValueError(
                    "actual values dataframe supplied but shape does not match original data"
                )

        if actual is None:
            warnings.warn(
                "Cannot score predictions on missing data because true values were never observed!"
            )
            return None

        with warnings.catch_warnings():
            # Catch 'Mean of empty slice' warnings from np.nanmean
            warnings.simplefilter("ignore", category=RuntimeWarning)
            if by_user:
                scores = []
                for userid in range(actual.shape[0]):
                    user_actual = actual.iloc[userid, :].values
                    user_pred = pred.iloc[userid, :].values
                    if metric == "rmse":
                        score = np.sqrt(np.nanmean((user_pred - user_actual) ** 2))
                    elif metric == "mse":
                        score = np.nanmean((user_pred - user_actual) ** 2)
                    elif metric == "mae":
                        score = np.nanmean(np.abs(user_pred - user_actual))
                    elif metric == "correlation":
                        nans = np.logical_or(np.isnan(user_actual), np.isnan(user_pred))
                        if len(user_actual[~nans]) < 2 or len(user_pred[~nans]) < 2:
                            score = np.nan
                        else:
                            score = pearsonr(user_actual[~nans], user_pred[~nans])[0]
                    scores.append(score)
                return pd.Series(scores, index=actual.index, name=f"{metric}_{dataset}")
            else:
                actual, pred = actual.to_numpy().flatten(), pred.to_numpy().flatten()

                if metric == "rmse":
                    return np.sqrt(np.nanmean((pred - actual) ** 2))
                elif metric == "mse":
                    return np.nanmean((pred - actual) ** 2)
                elif metric == "mae":
                    return np.nanmean(np.abs(pred - actual))
                elif metric == "correlation":
                    nans = np.logical_or(np.isnan(actual), np.isnan(pred))
                    if len(actual[~nans]) < 2 or len(pred[~nans]) < 2:
                        return np.nan
                    else:
                        return pearsonr(actual[~nans], pred[~nans])[0]

    def create_masked_data(self, n_mask_items=0.2):
        """
        Create a mask and apply it to data using number of items or % of items

        Args:
            n_items (int/float, optional): if an integer is passed its raw value is used. Otherwise if a float is passed its taken to be a (rounded) percentage of the total items; Default 0.1 (10% of the data)
        """

        if (
            isinstance(n_mask_items, np.floating)
            and (n_mask_items >= 1.0 or n_mask_items <= 0.0)
        ) or (
            isinstance(n_mask_items, int)
            and (n_mask_items >= self.data.shape[1] or n_mask_items <= 0)
        ):
            raise TypeError(
                "n_items should a float between 0-1 or an integer < the number of items"
            )
        self.mask = create_sparse_mask(
            self.data, n_mask_items, random_state=self.random_state
        )
        self.masked_data = self.data[self.mask]
        self.is_masked = True
        self.n_mask_items = n_mask_items

    def plot_predictions(
        self,
        dataset="missing",
        figsize=(16, 8),
        label_fontsize=16,
        hide_title=False,
        heatmap_kwargs={},
    ):
        """Create plot of actual vs predicted values.

        Args:
            dataset (str; optional): one of 'full', 'observed', or 'missing'. Default 'missing'.
            figsize (tuple; optional): matplotlib figure size; Default (16,8)
            label_fontsize (int; optional): fontsize for all axis labels and titles; Default 16
            hide_title (bool; optional): hide title containing RMSE and correlation performance if available; Default False
            heatmap_kwargs (dict, optional): addition arguments to seaborn.heatmap.

        Returns:
            tuple: (figure handle, axis handle)

        """

        if not self.is_fit:
            raise ValueError("Model has not been fit")

        vmax = max(self.data.max().max(), self.data.max().max())
        vmin = min(self.data.min().min(), self.data.min().min())

        actual, pred = self._retrieve_predictions(dataset)

        if actual is None:
            ncols = 2
            warnings.warn(
                "Cannot score predictions on missing data because true values were never observed!"
            )
        else:
            ncols = 3

        heatmap_kwargs.setdefault("square", False)
        heatmap_kwargs.setdefault("xticklabels", False)
        heatmap_kwargs.setdefault("yticklabels", False)
        heatmap_kwargs.setdefault("vmax", vmax)
        heatmap_kwargs.setdefault("vmin", vmin)

        f, ax = plt.subplots(nrows=1, ncols=ncols, figsize=figsize)

        # The original data matrix (potentially masked)
        sns.heatmap(self.masked_data, ax=ax[0], **heatmap_kwargs)
        ax[0].set_title("Actual User/Item Ratings", fontsize=label_fontsize)
        ax[0].set_xlabel("Items", fontsize=label_fontsize)
        ax[0].set_ylabel("Users", fontsize=label_fontsize)

        # The predicted data matrix
        sns.heatmap(self.predictions, ax=ax[1], **heatmap_kwargs)
        ax[1].set_title("Predicted User/Item Ratings", fontsize=label_fontsize)
        ax[1].set_xlabel("Items", fontsize=label_fontsize)
        ax[1].set_ylabel("Users", fontsize=label_fontsize)
        f.tight_layout()

        # Scatter plot if we can calculate it
        if actual is not None:
            nans = np.logical_or(np.isnan(actual), np.isnan(pred))
            ax[2].scatter(
                actual[~nans],
                pred[~nans],
            )
            ax[2].set_xlabel("Actual", fontsize=label_fontsize)
            ax[2].set_ylabel("Predicted", fontsize=label_fontsize)
            ax[2].set_title("Ratings", fontsize=label_fontsize)
            sns.despine()

            r = self.score(dataset=dataset, by_user=True, metric="correlation")
            rmse = self.score(dataset=dataset, by_user=True, metric="rmse")
            if not hide_title:
                plt.suptitle(
                    f"Mean RMSE: {np.round(rmse.mean(),3)} +/- {np.round(rmse.std(), 3)}\nMean Correlation: {np.round(r.mean(), 3)} +/- {np.round(r.std(), 3)}",
                    y=1.07,
                    fontsize=label_fontsize + 2,
                )
            plt.subplots_adjust(wspace=0.2)

        return f, ax

    def downsample(self, n_samples, sampling_freq=None, target_type="samples"):

        """
        Downsample a model's rating matrix to a new target frequency or number of samples using averaging. Also downsamples a model's mask and dilated mask if they exist as well as a model's predictions if it's already been fit.

        If target_type = 'samples' and sampling_freq is None, the new user x item matrix will have shape users x items * (1 / n_samples).

        If target_type = 'seconds', the new user x item matrix will have shape users x items * (1 / n_samples * sampling_freq).

        If target_type = 'hz', the new user x item matrix will have shape users x items * (1 / sampling_freq / n_samples).

        Args:
            n_samples (int): number of samples
            sampling_freq (int/float):  Sampling frequency of data; Default None
            target_type (str, optional): how to downsample; must be one of "samples", "seconds" or "hz". Defaults to "samples".

        """

        self.data = downsample_dataframe(
            self.data,
            sampling_freq=sampling_freq,
            n_samples=n_samples,
            target_type=target_type,
        )

        if self.is_masked:
            # Also downsample mask
            self.mask = downsample_dataframe(
                self.mask,
                sampling_freq=sampling_freq,
                n_samples=n_samples,
                target_type=target_type,
            )
            # Ensure mask stays boolean
            self.mask.loc[:, :] = self.mask > 0

            # Masked data
            self.masked_data = downsample_dataframe(
                self.masked_data,
                sampling_freq=sampling_freq,
                n_samples=n_samples,
                target_type=target_type,
            )
            # Dilated mask
            if self.is_mask_dilated:
                self.dilated_mask = downsample_dataframe(
                    self.dilated_mask,
                    sampling_freq=sampling_freq,
                    n_samples=n_samples,
                    target_type=target_type,
                )
                # Ensure mask stays boolean
                self.dilated_mask.loc[:, :] = self.dilated_mask > 0

        if self.is_fit:
            self.predictions = downsample_dataframe(
                self.predictions,
                sampling_freq=sampling_freq,
                n_samples=n_samples,
                target_type=target_type,
            )

    def to_long_df(self):

        """Create a long format pandas dataframe with observed, predicted, and mask."""

        observed = pd.DataFrame(columns=["User", "Item", "Rating", "Condition"])
        for row in self.data.iterrows():
            tmp = pd.DataFrame(columns=observed.columns)
            tmp["Rating"] = row[1]
            tmp["Item"] = self.data.columns
            tmp["User"] = row[0]
            tmp["Condition"] = "Observed"
            if self.is_masked:
                if self.is_mask_dilated:
                    tmp["Mask"] = self.dilated_mask.loc[row[0]]
                else:
                    tmp["Mask"] = self.mask.loc[row[0]]
            observed = observed.append(tmp)

        if self.is_fit:
            predicted = pd.DataFrame(columns=["User", "Item", "Rating", "Condition"])
            for row in self.predictions.iterrows():
                tmp = pd.DataFrame(columns=predicted.columns)
                tmp["Rating"] = row[1]
                tmp["Item"] = self.predictions.columns
                tmp["User"] = row[0]
                tmp["Condition"] = "Predicted"
                if self.is_masked:
                    tmp["Mask"] = self.mask.loc[row[0]]
                predicted = predicted.append(tmp)
            observed = observed.append(predicted)
        return observed

    def _retrieve_predictions(self, dataset):
        """Helper function to extract predicted values

        Args:
            dataset (str): should be one of 'full', 'observed', or 'missing''

        Returns:
            actual (array): true values
            predicted (array): predicted values
        """

        if dataset not in ["full", "observed", "missing"]:
            raise ValueError("dataset must be one of ['full', 'observed', 'missing']")

        if dataset == "full":
            return (self.data, self.predictions)

        # NOTE: always use self.mask to grab locations of observed and missing values rather than where masked_data is or isn't null. This is because dilation will by design *reduce* the sparsity of self.masked_data, meaning non-null indices are only a subset of the true missing values
        elif dataset == "observed":
            return (
                self.data[self.mask],
                self.predictions[self.mask],
            )
        elif dataset == "missing":
            if self.is_dense:
                return (
                    self.data[~self.mask],
                    self.predictions[~self.mask],
                )
            else:
                # This happens if the input data already has NaNs that were not the result of a masking operation by a model on dense data. In this case we never observed ground truth values to compare against predictions
                return (None, self.predictions[~self.mask])

    @staticmethod
    def _conv_ts_mean_overlap(sub_rating, n_samples=5):

        """Dilate each rating by n samples (centered).  If dilated samples are overlapping they will be averaged.

        Args:
            sub_rating (array): vector of data for subject
            n_samples (int):  number of samples to dilate each rating

        Returns:
            sub_rating_conv_mn (array): subject rating vector with each rating dilated n_samples (centered) with mean of overlapping

        """

        # Notes:  Could add custom filter input
        bin_sub_rating = ~sub_rating.isnull()
        if np.any(sub_rating.isnull()):
            sub_rating.fillna(0, inplace=True)
        filt = np.ones(n_samples)
        bin_sub_rating_conv = np.convolve(bin_sub_rating, filt)[: len(bin_sub_rating)]
        sub_rating_conv = np.convolve(sub_rating, filt)[: len(sub_rating)]
        sub_rating_conv_mn = deepcopy(sub_rating_conv)
        sub_rating_conv_mn[bin_sub_rating_conv >= 1] = (
            sub_rating_conv_mn[bin_sub_rating_conv >= 1]
            / bin_sub_rating_conv[bin_sub_rating_conv >= 1]
        )
        new_mask = bin_sub_rating_conv == 0
        sub_rating_conv_mn[new_mask] = np.nan
        return sub_rating_conv_mn

    def dilate_mask(self, n_samples=None):

        """Dilate sparse time-series data by n_samples.
        Overlapping data will be averaged. This method computes and stores the dilated mask in `.dilated_mask` and internally updates the `.masked_data`. Repeated calls to this method on the same model instance **do not** stack, but rather perform a new dilation on the original masked data. Called this method with `None` will undo any dilation.

        Args:
            nsamples (int):  Number of samples to dilate data

        """

        if self.mask is None:
            raise ValueError("Model has no mask and requires one to perform dilation")
        if not self.is_masked and n_samples is not None:
            raise ValueError("Make sure model instance has been masked.")

        if isinstance(n_samples, np.floating) or (
            n_samples is not None and n_samples >= self.data.shape[1]
        ):
            raise TypeError("nsamples should be an integer < the number of items")

        # Always reset to the undilated mask first
        self.masked_data = self.data[self.mask]

        if n_samples is not None:
            # After masking, perform dilation and save as the new masked data
            self.masked_data = self.masked_data.apply(
                lambda x: self._conv_ts_mean_overlap(x, n_samples=n_samples),
                axis=1,
                result_type="broadcast",
            )
            # Calculate and save dilated mask
            self.dilated_mask = ~self.masked_data.isnull()
            self.is_mask_dilated = True
            self.dilated_by_nsamples = n_samples
        else:
            self.dilated_mask = None
            self.is_mask_dilated = False
            self.dilated_by_nsamples = None

    def fit(self, **kwargs):
        """Replaced by sub-classes. This call just ensures that a model's data is sparse prior to fitting"""
        if not self.is_masked:
            raise ValueError(
                "You're trying to fit on a dense matrix, because model data has not been masked! Either call the `.create_masked_data` method prior to fitting or re-initialize the model and set the `mask` or `n_mask_items` arguments."
            )
        if kwargs.get("dilate_by_nsamples", None) and self.is_mask_dilated:
            warnings.warn(
                ".fit() was called with dilate_by_nsamples=None, but model mask is already dilated! This will undo dilation and then fit a model. Instead pass dilate_by_nsamples, directly to .fit()"
            )

    def summary(self, verbose=False, actual=None, dataset=None):
        """
        Calculate the performance of a model and return a dataframe of results. Computes performance across all, observed, and missing datasets. Scores using rmse, mse, mae, and correlation. Computes scores across all subjects (i.e. ignoring the fact that ratings are clustered by subject) and the mean performance for each metric after calculating per-subject performance.

        Args:
            verbose (bool, optional): Print warning messages during scoring. Defaults to False.
            actual (pd.DataFrame, None; optional): a dataframe to score against; Default is None which uses the data provided when the model was initialized
            dataset (str/None): dataset to score. Must be one of 'full', 'observed','missing' or None to score both 'observed' and 'missing'; Default None

        Returns:
            pd.DataFrame: long-form dataframe of model performance
        """

        if not self.is_fit:
            raise ValueError("Model has not been fit!")

        if dataset is None:
            if actual is None:
                if self.is_dense:
                    dataset = ["missing", "observed"]
                else:
                    dataset = ["observed"]
            else:
                dataset = ["missing", "observed"]

        elif isinstance(dataset, str):
            if actual is None and not self.is_dense and dataset in ["full", "missing"]:
                raise ValueError(
                    "Cannot score predictions on missing values because no ground truth was observed"
                )
            dataset = [dataset]

        # Compute results for all metrics, all datasets, separately for group and by subject
        group_results = {
            "algorithm": self.__class__.__name__,
        }
        subject_results = []

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            for metric in ["rmse", "mse", "mae", "correlation"]:
                this_group_result = {}
                this_subject_result = []
                for dat in dataset:
                    this_group_result[dat] = self.score(
                        metric=metric, dataset=dat, actual=actual, by_user=False
                    )
                    this_subject_result.append(
                        self.score(
                            metric=metric,
                            dataset=dat,
                            by_user=True,
                            actual=actual,
                        )
                    )
                # Dict of group results for this metric
                group_results[metric] = this_group_result
                # Dataframe of subject results for this metric
                this_subject_result = pd.concat(this_subject_result, axis=1)
                subject_results.append(this_subject_result)
                group_results[f"{metric}_user"] = dict(
                    zip(
                        dataset,
                        this_subject_result.mean().values,
                    )
                )
        # Save final results to longform df
        self.user_results = pd.concat(subject_results, axis=1)
        group_results = pd.DataFrame(group_results)
        group_results = (
            group_results.reset_index()
            .melt(
                id_vars=["index", "algorithm"],
                var_name="metric",
                value_name="score",
            )
            .rename(columns={"index": "dataset"})
            .sort_values(by=["dataset", "metric"])
            .reset_index(drop=True)
            .assign(
                group=lambda df: df.metric.apply(
                    lambda x: "user" if "user" in x else "all"
                ),
                metric=lambda df: df.metric.replace(
                    {
                        "correlation_user": "correlation",
                        "mse_user": "mse",
                        "rmse_user": "rmse",
                        "mae_user": "mae",
                    }
                ),
            )
            .sort_values(by=["dataset", "group", "metric"])
            .reset_index(drop=True)[
                ["algorithm", "dataset", "group", "metric", "score"]
            ]
        )
        self.overall_results = group_results
        if verbose:
            if w:
                print(w[-1].message)
            print(
                "User performance results (not returned) are accessible using .user_results"
            )
            print(
                "Overall performance results (returned) are accesible using .overall_results"
            )

        return group_results

    def transform(self, return_only_predictions=False):
        """
        Return a user x item matrix of predictions after a model has been fit

        Args:
            return_only_predictions (bool, optional): Returns both training and testing predictions rather than simply filling in missing values with predictions. Defaults to False.

        Returns:
            pd.DataFrame: user x item ratings
        """

        if not self.is_fit:
            raise ValueError("Model has not been fit!")
        if return_only_predictions:
            return self.predictions
        else:
            # Propagate observed values to return object
            out = self.data[self.mask]
            # Fill in missing values with predictions
            out[~self.mask] = self.predictions[~self.mask]
            return out


class BaseNMF(Base):
    """
    Base class for NMF algorithms.
    """

    def __init__(
        self, data, mask=None, n_mask_items=None, verbose=True, random_state=None
    ):
        super().__init__(
            data, mask, n_mask_items, random_state=random_state, verbose=verbose
        )
        self.error_history = []

    def plot_learning(self, save=False):
        """
        Plot training error over iterations for diagnostic purposes

        Args:
            save (bool/str/Path, optional): if a string or path is provided will save the figure to that location. Defaults to False.

        Returns:
            tuple: (figure handle, axes handle)
        """

        if self.is_fit:
            f, ax = plt.subplots(1, 1, figsize=(8, 6))
            _ = ax.plot(range(1, len(self.error_history) + 1), self.error_history)
            ax.set(
                xlabel="Iteration",
                ylabel="Normalized RMSE",
                title=f"Final Normalized RMSE: {np.round(self._norm_rmse, 3)}\nConverged: {self.converged}",
            )
            sns.despine()
            if save:
                plt.savefig(save, bbox_inches="tight")
            return f, ax
        else:
            raise ValueError("Model has not been fit.")

    def plot_factors(self, save=False, **kwargs):
        """
        Plot user x factor and item x factor matrices

        Args:
            save (bool/str/Path, optional): if a string or path is provided will save the figure to that location. Defaults to False.
            kwargs: additional arguments to seaborn.heatmap

        Returns:
            tuple: (figure handle, axes handle)
        """

        if self.is_fit:
            f, axs = plt.subplots(1, 2, figsize=(12, 6))
            if hasattr(self, "W"):
                _ = sns.heatmap(self.W, ax=axs[0], **kwargs)
                _ = sns.heatmap(self.H, ax=axs[1], **kwargs)
            else:
                _ = sns.heatmap(self.user_vecs, ax=axs[0], **kwargs)
                _ = sns.heatmap(self.item_vecs, ax=axs[1], **kwargs)
            axs[0].set_xlabel("Factor", fontsize=18)
            axs[0].set_ylabel("User", fontsize=18)
            axs[0].set_title("User Factors", fontsize=18)
            axs[1].set_xlabel("Item", fontsize=18)
            axs[1].set_ylabel("Factor", fontsize=18)
            axs[1].set_title("Item Factors", fontsize=18)
            if save:
                plt.savefig(save, bbox_inches="tight")
            return f, axs
        else:
            raise ValueError("Model has not been fit.")



from setuptools import setup, find_packages

version = {}
with open("neighbors/version.py") as f:
    exec(f.read(), version)

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

extra_setuptools_args = dict(tests_require=["pytest"])

setup(
    name="neighbors",
    version=version["__version__"],
    description="A Python package for performing collaborative filtering on social and emotion datasets",
    maintainer="Cosan Lab",
    maintainer_email="eshin.jolly@dartmouth.edu",
    url="http://github.com/cosanlab/neighbors",
    install_requires=requirements,
    packages=find_packages(exclude=["neighbors/tests"]),
    include_package_data=True,
    package_data={"": ["data/*.csv"]},
    license="MIT",
    keywords=["emotion", "collaborative filtering", "recommender", "machine-learning"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
    ],
    **extra_setuptools_args
)



"""
Module that holds functions supporting various model methods. Not designed to be user-facing
"""

import numpy as np
import numba as nb


@nb.njit(cache=True, nogil=True)
def sgd(
    data,
    seed,
    global_bias,
    data_range,
    tol,
    user_bias,
    user_vecs,
    user_bias_reg,
    user_fact_reg,
    item_bias,
    item_vecs,
    item_bias_reg,
    item_fact_reg,
    n_iterations,
    row_indices,
    col_indices,
    learning_rate,
    verbose,
):
    """SGD Update. This implementation is nearly identitical the the SVD implementation used by Simon Funk in the Netflix challenge and implemented in Surprise with a few small differences. We currently only support a single learning rate for all parameters (Surprise supports independent learning rates, but doesn't use them by default), we don't train in batched epochs but rather over *all* training data in each iteration, and we force user and item factor values to be >=0 after each pass over the training data."""

    error_history = np.zeros((n_iterations))
    converged = False
    last_e = 0
    norm_rmse = np.inf
    delta = np.inf
    np.random.seed(seed)
    for this_iter in range(n_iterations):

        # Generate shuffled order to loop over training data
        # Recall that row_indices and col_indices need to be looped over simultaneously to properly index each training value at [row, col]
        training_indices = np.arange(len(row_indices))
        np.random.shuffle(training_indices)

        if verbose and this_iter > 0 and this_iter % 10 == 0:
            disp_norm_error = np.round(100 * norm_rmse, 2)
            # Numba doesn't like f-strings
            print(
                "Iter:",
                this_iter,
                " Norm RMSE:",
                disp_norm_error,
                "%  Delta Convg:",
                delta,
                "||",
                tol,
            )

        # Loop over every data point in the training set, make a prediction, calculate error, and update biases and vectors
        # Very similar to Surprise's SVD implementation: https://github.com/NicolasHug/Surprise/blob/master/surprise/prediction_algorithms/matrix_factorization.pyx#L159
        # Because we're iterating an user-item combo at a time, track total error
        total_error = 0
        for idx in training_indices:
            u = row_indices[idx]
            i = col_indices[idx]

            prediction = global_bias + user_bias[u] + item_bias[i]
            prediction += user_vecs[u, :] @ item_vecs[:, i]

            # Use changes in e to determine tolerance
            e = data[u, i] - prediction  # error

            # Update biases
            user_bias[u] += learning_rate * (e - user_bias_reg * user_bias[u])
            item_bias[i] += learning_rate * (e - item_bias_reg * item_bias[i])

            # Update latent factors
            user_vecs[u, :] += learning_rate * (
                e * item_vecs[:, i] - user_fact_reg * user_vecs[u, :]
            )
            item_vecs[:, i] += learning_rate * (
                e * user_vecs[u, :] - item_fact_reg * item_vecs[:, i]
            )

            # Keep track of total squared error
            total_error += np.power(e, 2)

        # Force non-negativity. Surprise does this per-epoch via re-initialization. We do this per sweep over all training data, e.g. see: https://github.com/NicolasHug/Surprise/blob/master/surprise/prediction_algorithms/matrix_factorization.pyx#L671
        user_vecs = np.maximum(user_vecs, 0)
        item_vecs = np.maximum(item_vecs, 0)

        # Normalize the current error with respect to the range of the dataset
        rmse = np.sqrt(total_error / len(training_indices))
        norm_rmse = rmse / data_range
        error_history[this_iter] = norm_rmse

        # Compute the delta to see if we should stop iterating
        delta = np.abs(np.abs(norm_rmse) - np.abs(last_e))
        # If we've converged break out of iteration
        if delta < tol:
            converged = True
            error_history = error_history[: this_iter + 1]
            break
        # Otherwise update the error
        last_e = norm_rmse

    return (
        error_history,
        converged,
        this_iter,
        delta,
        norm_rmse,
        user_bias,
        user_vecs,
        item_bias,
        item_vecs,
    )


@nb.njit(cache=True, nogil=True)
def mult(X, W, H, data_range, eps, tol, n_iterations, verbose):
    """Lee & Seung (2001) multiplicative update rule"""

    last_e = 0
    error_history = np.zeros((n_iterations))
    converged = False
    norm_rmse = np.inf
    delta = np.inf

    for this_iter in range(n_iterations):

        if verbose and this_iter > 0 and this_iter % 10 == 0:
            disp_norm_error = np.round(100 * norm_rmse, 2)
            # Numba doesn't like f-strings
            print(
                "Iter:",
                this_iter,
                " Norm RMSE:",
                disp_norm_error,
                "%  Delta Convg:",
                delta,
                "||",
                tol,
            )

        # Update H
        numer = W.T @ X
        denom = W.T @ W @ H + eps
        H *= numer
        H /= denom

        # Update W
        numer = X @ H.T
        denom = W @ H @ H.T + eps
        W *= numer
        W /= denom

        # Make prediction and get error
        errors = X - W @ H
        rmse = np.sqrt(np.mean(np.power(errors, 2)))

        # Normalize current error with respect to max of dataset
        norm_rmse = rmse / data_range
        error_history[this_iter] = norm_rmse

        # Compute delta to see if we should stop iterating
        delta = np.abs(np.abs(norm_rmse) - np.abs(last_e))
        if delta < tol:
            converged = True
            error_history = error_history[: this_iter + 1]
            break
        # Otherwise update the error
        last_e = norm_rmse

    return error_history, converged, this_iter, delta, norm_rmse, W, H



"""
Test utility functions
"""

import numpy as np
import pandas as pd
from neighbors import (
    create_user_item_matrix,
    invert_user_item_matrix,
    nanpdist,
    create_sparse_mask,
    estimate_performance,
    flatten_dataframe,
    unflatten_dataframe,
    split_train_test,
    Mean,
)


def test_estimate_performance(simulate_wide_data):
    # DENSE DATA
    # 1) Only out-of-sample
    group_out, user_out = estimate_performance(Mean, simulate_wide_data, verbose=True)
    assert group_out.shape == (4 * 2, 6)
    assert user_out.shape == (50, 4)
    # 2) In and out-of-sample
    group_out, user_out = estimate_performance(
        Mean,
        simulate_wide_data,
        verbose=True,
        return_full_performance=True,
        parallelize=True,
    )
    assert group_out.shape == (4 * 2 * 2, 6)
    assert user_out.shape == (50, 8)
    missing = group_out.query("dataset == 'missing' and group =='all'")["mean"]
    observed = group_out.query("dataset == 'observed' and group =='all'")["mean"]
    # Make sure all missing scores are worse than observed
    for i, (o, m) in enumerate(zip(observed, missing)):
        if i == 0:
            assert o > m
        else:
            assert o < m
    # 3) Non-aggregated out-of-sample, with timing info
    group_out, user_out = estimate_performance(
        Mean, simulate_wide_data, return_agg=False, verbose=True
    )
    assert group_out.shape == (4 * 2 * 10, 7)
    assert user_out.shape == (50 * 10, 6)

    # SPARSE DATA
    mask = create_sparse_mask(simulate_wide_data)
    masked_data = simulate_wide_data[mask]
    # 1) Only out-of-sample
    group_out, user_out = estimate_performance(Mean, masked_data, verbose=True)
    assert group_out.shape == (4 * 2, 6)
    assert user_out.shape == (50, 4)
    # 2) In and out-of-sample
    group_out, user_out = estimate_performance(
        Mean,
        simulate_wide_data,
        verbose=True,
        return_full_performance=True,
        parallelize=True,
    )
    assert group_out.shape == (4 * 2 * 2, 6)
    assert user_out.shape == (50, 8)
    test = group_out.query("dataset == 'test' and group =='all'")["mean"]
    train = group_out.query("dataset == 'train' and group =='all'")["mean"]
    # Make sure all test scores are worse than train
    for i, (o, m) in enumerate(zip(train, test)):
        if i == 0:
            assert o > m
        else:
            assert o < m
    # 3) Non-aggregated out-of-sample, without timing info
    group_out, user_out = estimate_performance(
        Mean, masked_data, return_agg=False, verbose=True, timeit=False
    )
    assert group_out.shape == (4 * 2 * 10, 6)
    assert user_out.shape == (50 * 10, 6)


# TODO: Update this test to handle commented out lines. This is a pandas issue where going from long -> wide -> long leads pandas to sort columns rather than preserving the original column order
def test_create_and_invert_user_item_matrix(simulate_long_data, simulate_wide_data):
    rating = create_user_item_matrix(simulate_long_data)
    assert isinstance(rating, pd.DataFrame)
    assert rating.shape == (50, 100)
    # assert rating.equals(simulate_wide_data)

    inverted = invert_user_item_matrix(rating)
    assert inverted.shape == (50 * 100, 3)
    # assert inverted.equals(simulate_long_data)

    renamed = simulate_long_data.rename(
        columns={"User": "A", "Item": "B", "Rating": "C"}
    )
    rating = create_user_item_matrix(renamed, columns=["A", "B", "C"])
    assert isinstance(rating, pd.DataFrame)
    assert rating.shape == (50, 100)


def test_nanpdist(simulate_wide_data):
    # Non-nan data should behave like pdist
    out = nanpdist(simulate_wide_data.to_numpy())
    assert out.ndim == 2
    assert np.allclose(out, out.T, rtol=1e-05, atol=1e-08)
    out = nanpdist(simulate_wide_data.to_numpy(), return_square=False)
    assert out.ndim == 1

    # Now mask it
    mask = np.random.choice([0, 1], size=simulate_wide_data.shape)
    df = simulate_wide_data * mask
    df = df.replace({0: np.nan})

    out = nanpdist(df.to_numpy())
    assert out.ndim == 2
    assert np.allclose(out, out.T, rtol=1e-05, atol=1e-08)
    out = nanpdist(df.to_numpy(), return_square=False)
    assert out.ndim == 1

    calc_corr_mat = 1 - nanpdist(df.to_numpy(), metric="correlation")
    pd_corr_mat = df.T.corr(method="pearson").to_numpy()
    assert np.allclose(calc_corr_mat, pd_corr_mat)


def test_create_sparse_mask(simulate_wide_data):
    mask = create_sparse_mask(simulate_wide_data, n_mask_items=0.1)
    expected_items = int(simulate_wide_data.shape[1] * (1 - 0.10))
    assert mask.shape == simulate_wide_data.shape
    assert all(mask.sum(1) == expected_items)

    mask = create_sparse_mask(simulate_wide_data, n_mask_items=19)
    assert mask.shape == simulate_wide_data.shape
    expected_items = int(simulate_wide_data.shape[1] - 19)
    assert all(mask.sum(1) == expected_items)

    masked_data = simulate_wide_data[mask]
    assert isinstance(masked_data, pd.DataFrame)
    assert masked_data.shape == simulate_wide_data.shape
    assert ~simulate_wide_data.isnull().any().any()
    assert masked_data.isnull().any().any()


def test_flatten_dataframe(simulate_wide_data):
    out = flatten_dataframe(simulate_wide_data)
    assert isinstance(out, np.ndarray)
    assert len(out) == simulate_wide_data.shape[0] * simulate_wide_data.shape[1]
    assert all([len(elem) == 3 for elem in out])


def test_unflatten_dataframe(simulate_wide_data):
    out = flatten_dataframe(simulate_wide_data)
    new = unflatten_dataframe(
        out, index=simulate_wide_data.index, columns=simulate_wide_data.columns
    )
    assert new.equals(simulate_wide_data)
    new = unflatten_dataframe(out)
    assert new.equals(simulate_wide_data)
    new = unflatten_dataframe(
        out, num_rows=simulate_wide_data.shape[0], num_cols=simulate_wide_data.shape[1]
    )


def test_split_train_test(simulate_wide_data):
    # Dense data
    for train, test in split_train_test(simulate_wide_data, n_folds=5):
        assert train.shape == simulate_wide_data.shape
        assert test.shape == simulate_wide_data.shape
        # 1/5 of dense data should be sparse for training, i.e. 4/5 data folds
        assert train.notnull().sum().sum() == int(4 / 5 * simulate_wide_data.size)
        # 4/5 of dense data should be sparse for testing, i.e. 1/5 data folds
        assert test.notnull().sum().sum() == int(1 / 5 * simulate_wide_data.size)
        assert train.add(test, fill_value=0).equals(simulate_wide_data)

    # Sparse data
    mask = create_sparse_mask(simulate_wide_data, n_mask_items=0.1)
    masked_data = simulate_wide_data[mask]
    masked_not_null = masked_data.notnull().sum().sum()

    for train, test in split_train_test(masked_data, n_folds=5):
        assert train.shape == mask.shape
        assert test.shape == mask.shape
        # Train and test should be more sparse than original
        assert train.isnull().sum().sum() > masked_data.isnull().sum().sum()
        assert test.isnull().sum().sum() > masked_data.isnull().sum().sum()
        assert train.add(test, fill_value=0).equals(masked_data)
        train_not_null = train.notnull().sum().sum()
        test_not_null = test.notnull().sum().sum()
        # And adhere close the expected train/test split
        assert np.allclose(train_not_null / masked_not_null, 0.8, atol=0.12)
        assert np.allclose(test_not_null / masked_not_null, 0.2, atol=0.12)
